/*  sound.c
 *
 *  Soundcard handling functions of demorse application
 */

/*
 *  demorse: An application to decode Morse code signals to text
 *
 *  Copyright (C) 2002  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */


#include "demorse.h"

extern int
  dsp_fd,   /* File descriptor of dsp device */
  mix_fd;   /* File descriptor of mixer device */

extern rc_data_t rc_data;

/*------------------------------------------------------------------------*/

/*  Setup_Sound_Card()
 *
 *  Sets up mixer and dsp devices
 */

  void
Setup_Sound_Card( void )
{
  int
	itmp,       /* Temporary integer for tests etc   */
	new_recsrc, /* New recording source (RECORD_SRC) */
	rec_src,    /* Actual recording source for tests */
	dsp_speed,  /* dsp sampling speed (9600 Hz)      */
	frag_size,  /* dsp fragment (chunk) size         */
	record_lev, /* Recording level request to mixer  */
	channels,   /* Stereo(=2) or Mono(=1) selection  */
	sample_fmt; /* dsp sample format, 8-bit unsigned */

  /* Mixer device to be set-up (input or reclev) */
  unsigned int device;

  /* Sound device names from souncard.h */
  char *dev_name[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_NAMES;

  /* File lock information   */
  struct flock lockinfo;

  /* Open mixer device, abort on error */
  if( (mix_fd = open(rc_data.mix_dev, O_RDONLY, 0)) == -1 )
  {
	perror( "demorse: open(rc_data.mix_dev, O_RDONLY, 0)" );
	fprintf( stderr, "demorse: failed to open mixer device: %s%s\n",
		rc_data.mix_dev, " - aborting" );
	exit(-1);
  }

  /* Attempt to lock mixer device */
  fprintf( stderr, "demorse: attempting to lock %s%s\n",
	  rc_data.mix_dev, " ..." );

  /* Attempt to lock entire mixer device file */
  lockinfo.l_type   = F_RDLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If mixer device is already locked, abort */
  if( fcntl( mix_fd, F_SETLK, &lockinfo ) < 0 )
  {
	perror( "demorse: fcntl( mix_fd, F_SETLK, &lockinfo )" );
	fcntl( mix_fd, F_GETLK, &lockinfo );
	fprintf( stderr, "demorse: lock mixer device: %s\n"
		": device already locked by pid %d%s\n",
		rc_data.mix_dev, lockinfo.l_pid, " - aborting" );
	exit( -1 );
  }

  /* Open dsp device, abort on error */
  if( (dsp_fd = open(rc_data.dsp_dev, O_RDONLY, 0)) == -1 )
  {
	perror( "demorse: open(rc_data.dsp_dev, O_RDONLY, 0)" );
	fprintf( stderr, "demorse: failed to open dsp device: %s%s\n",
		rc_data.dsp_dev, " - aborting" );
	exit(-1);
  }

  /* Attempt to lock dsp device */
  fprintf( stderr, "demorse: attempting to lock %s%s\n",
	  rc_data.dsp_dev, " ..." );

  /* Attempt to lock entire dsp device file */
  lockinfo.l_type   = F_RDLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If dsp device is already locked, abort */
  if( fcntl( dsp_fd, F_SETLK, &lockinfo ) < 0 )
  {
	perror( "demorse: fcntl( dsp_fd, F_SETLK, &lockinfo )" );
	fcntl( dsp_fd, F_GETLK, &lockinfo );
	fprintf( stderr, "demorse: lock dsp device: %s\n"
		": device already locked by pid %d%s\n",
		rc_data.dsp_dev, lockinfo.l_pid, " - aborting" );
	exit( -1 );
  }

  /* Attempt to set up sound card */
  fprintf( stderr, "%s\n", "demorse: setting up sound card ..." );

  /* Find number of recording source matching source name */
  for( new_recsrc = 0; new_recsrc < SOUND_MIXER_NRDEVICES; new_recsrc++ )
	if( strcmp( dev_name[new_recsrc], rc_data.rec_src ) == 0 )
	  break;

  /* Write recording source level to mixer, abort on error */
  record_lev = itmp = rc_data.rec_lev | (rc_data.rec_lev << 8);
  if( ioctl(mix_fd, MIXER_WRITE(new_recsrc), &record_lev) == -1 )
  {
	perror("demorse: ioctl(mix_fd, MIXER_WRITE(new_recsrc), &record_lev)");
	fprintf( stderr, "demorse: setting recording level failed\n" );
	exit(-1);
  }
  if( record_lev != itmp )
	fprintf( stderr, "demorse: setting recording level failed\n"
		"         Try manually using a mixer application\n" );

  /* Make recording source mask */
  new_recsrc = itmp = 1 << new_recsrc;
  /* Attempt to select recording source, abort on error */
  if( ioctl(mix_fd, SOUND_MIXER_WRITE_RECSRC, &new_recsrc) == -1 )
  {
	perror("demorse: ioctl(mix_fd, SOUND_MIXER_WRITE_RECSRC, &new_recsrc)");
	fprintf( stderr, "demorse: failed to select recording source: %s%s\n",
		rc_data.rec_src, " - aborting" );
	exit(-1);
  }

  /* Check that recording source is correctly selected */
  if( ioctl(mix_fd, SOUND_MIXER_READ_RECSRC, &rec_src) == -1 )
  {
	perror("demorse: ioctl(mix_fd, SOUND_MIXER_READ_RECSRC, &rec_src)");
	fprintf( stderr, "demorse: failed to verify recording source: %s%s\n",
		rc_data.rec_src, " - aborting" );
	exit(-1);
  }
  if( rec_src != itmp )
	Abort_On_Error( -2 );

  /* Set fragment size according to stereo/mono mode */
  itmp = 1; frag_size = 0;
  channels = NUM_CHANNELS;
  while( (itmp << frag_size) != BUFFER_SIZE ) frag_size++;
  itmp += channels << 18;
  ioctl( dsp_fd, SNDCTL_DSP_SETFRAGMENT, &frag_size );

  /* Set sample size to unsigned 8-bit, abort on error */
  sample_fmt = AFMT_U8;
  if( ioctl(dsp_fd, SNDCTL_DSP_SETFMT, &sample_fmt) == -1 )
  {
	perror( "demorse: ioctl(dsp_fd, SNDCTL_DSP_SETFMT, &sample_fmt)");
	exit(-1);
  }
  if( sample_fmt != AFMT_U8 )
	Abort_On_Error( -3 );

  /* Set stereo or mono mode */
  channels = NUM_CHANNELS;
  if( ioctl(dsp_fd, SNDCTL_DSP_CHANNELS, &channels) == -1 )
  {
	perror("demorse: ioctl(dsp_fd, SNDCTL_DSP_CHANNELS, &channels)");
	exit(-1);
  }

  /* Setup dsp sampling speed, abort on error */
  dsp_speed = DSP_SPEED;
  if( ioctl(dsp_fd, SNDCTL_DSP_SPEED, &dsp_speed) == -1 )
  {
	perror( "demorse: ioctl(dsp_fd, SNDCTL_DSP_SPEED, &dsp_speed)");
	exit(-1);
  }
  if( dsp_speed != DSP_SPEED )
	Abort_On_Error( -4 );

  /* Find recording level device */
  for( device = 0; device < SOUND_MIXER_NRDEVICES; device++ )
	if( strcmp( dev_name[device], rc_data.inp_lev ) == 0 )
	  break;
  if( device ==  SOUND_MIXER_NRDEVICES )
  {
	fprintf( stderr, "demorse: failed to select recording level device: %s\n",
		rc_data.inp_lev );
	exit(-1);
  }

  /* Write recording level to mixer, abort on error */
  record_lev = itmp = rc_data.rec_lev | (rc_data.rec_lev << 8);
  if( ioctl(mix_fd, MIXER_WRITE(device), &record_lev) == -1 )
  {
	perror("demorse: ioctl(mix_fd, MIXER_WRITE(device), &record_lev");
	fprintf( stderr, "demorse: failed to set recording level: %d%s\n",
		rc_data.rec_lev, " - aborting" );
	exit(-1);
  }
  if( record_lev != itmp )
	fprintf( stderr, "demorse: setting recording level failed\n"
		"         Try manually using a mixer application\n" );


  printf( "%s\n", "demorse: sound card setup completed\n" );

} /* End of Setup_Sound_Card() */

/*------------------------------------------------------------------------*/


/*  Get_Signal_Sample()
 *
 *  Gets the next DSP sample of the signal input
 */

  int
Get_Signal_Sample(void)
{
  int sample_cnt;  /* Number of samples read from dsp */

  /* Signal/dsp samples buffer */
  static unsigned char buffer[BUFFER_SIZE * NUM_CHANNELS];

  static int
	/* Index to signal samples buffer */
	buffer_idx = BUFFER_SIZE * NUM_CHANNELS,
    /* Buffer size according to stereo/mono mode */
	buffer_size = BUFFER_SIZE * NUM_CHANNELS;

  /* New DSP sample of audio input  */
  int sample;


  /* Refill dsp samples buffer when needed */
  if( buffer_idx >= buffer_size )
  {
	/* Start buffer index according to stereo/mono mode */
	if( NUM_CHANNELS == 1 ) /* Mono */
	  buffer_idx = 0;
	else
	  buffer_idx = rc_data.use_chn;

	/* Read carrier samples from dsp, abort on error */
	if( (sample_cnt = read(dsp_fd, buffer, buffer_size)) == -1 )
	{
	  perror( "demorse: dsp read()" );
	  exit(-1);
	}
	if( sample_cnt != buffer_size )
	  Abort_On_Error( -6 );

  } /* End of if( buffer_idx == buffer_size ) */

  /* Remove neutral value (128) from samples */
  sample = buffer[ buffer_idx ] - 128;

  /* Increment according to mono/stereo mode */
  buffer_idx += NUM_CHANNELS;

  return( sample );

} /* End of Get_Signal_Sample() */

/*------------------------------------------------------------------------*/

/*  Setup_Signal_Input()
 *
 *  Displays information needed for the setup of
 *  the Morse sound signals fed to the sound card
 */

  void
Setup_Signal_Input( void )
{
  int
	/* Size of samples buffer depends on stereo/mono mode */
	buffer_size = BUFFER_SIZE * NUM_CHANNELS,
	sample_lev,  /* Normalized DSP sound sample level */
	sample_max,  /* Max value of DSP sound samples    */
	sample_cnt,  /* Number of samples read from DSP   */
	idx;         /* Index for loops etc */

  /* Signal/DSP samples buffer */
  unsigned char buffer[BUFFER_SIZE * NUM_CHANNELS];

  /* Buffer for graph plotting */
  char graph[ 46 ];

  while( 1 )
  {
	sample_max = 0;

	/* Read audio samples from DSP, abort on error */
	if( (sample_cnt = read(dsp_fd, buffer, buffer_size)) == -1 )
	{
	  perror( "demorse: dsp read()" );
	  exit(-1);
	}
	if( sample_cnt != buffer_size )
	  exit(0);

	/*** Find max sample level (average of 4 samples) ***/

	/* Start buffer index according to stereo/mono mode */
	if( NUM_CHANNELS == 1 ) /* Mono */
	  idx = 0;
	else
	  idx = rc_data.use_chn;

	for( ; idx < buffer_size; idx += NUM_CHANNELS )
	{
	  sample_lev = buffer[ idx ] - 128;

	  if( sample_max < sample_lev )
		sample_max = ( (sample_max * 3) + sample_lev ) / 4;
	}
	sample_max += 2;

	/* Make a graph of detector o/p level */
	for( idx = 0; idx < sample_max / 3; idx++ )
	  graph[ idx ] = '=';
	graph[ idx++ ] = '>';
	for( ; idx < 44; idx++ )
	  graph[ idx ] = '.';
	graph[ idx ] = '<';
	graph[ 45 ]  = '\0';

	fprintf( stderr, "%03d %s\n", sample_max, graph );

  } /* while( 1 ) */

} /* End of Setup_Signal_Input() */

/*------------------------------------------------------------------------*/
